@@ -77,7 +77,6 @@ void parseExpectErrorMeta(
 
   Expected<std::unique_ptr<remarks::RemarkParser>> MaybeParser =
       remarks::createRemarkParserFromMeta(remarks::Format::YAML, Buf,
-                                          /*StrTab=*/std::nullopt,
                                           std::move(ExternalFilePrependPath));
   handleAllErrors(MaybeParser.takeError(),
                   [&](const ErrorInfoBase &EIB) { EIB.log(Stream); });
@@ -558,124 +557,6 @@ TEST(YAMLRemarks, ContentsCAPI) {
   LLVMRemarkParserDispose(Parser);
 }
 
-TEST(YAMLRemarks, ContentsStrTab) {
-  StringRef Buf = "\n"
-                  "--- !Missed\n"
-                  "Pass: 0\n"
-                  "Name: 1\n"
-                  "DebugLoc: { File: 2, Line: 3, Column: 12 }\n"
-                  "Function: 3\n"
-                  "Hotness: 4\n"
-                  "Args:\n"
-                  "  - Callee: 5\n"
-                  "  - String: 7\n"
-                  "  - Caller: 3\n"
-                  "    DebugLoc: { File: 2, Line: 2, Column: 0 }\n"
-                  "  - String: 8\n"
-                  "\n";
-
-  StringRef StrTabBuf =
-      StringRef("inline\0NoDefinition\0file.c\0foo\0Callee\0bar\0String\0 "
-                "will not be inlined into \0 because its definition is "
-                "unavailable",
-                115);
-
-  remarks::ParsedStringTable StrTab(StrTabBuf);
-  Expected<std::unique_ptr<remarks::RemarkParser>> MaybeParser =
-      remarks::createRemarkParser(remarks::Format::YAMLStrTab, Buf,
-                                  std::move(StrTab));
-  EXPECT_FALSE(errorToBool(MaybeParser.takeError()));
-  EXPECT_TRUE(*MaybeParser != nullptr);
-
-  remarks::RemarkParser &Parser = **MaybeParser;
-  Expected<std::unique_ptr<remarks::Remark>> MaybeRemark = Parser.next();
-  EXPECT_FALSE(
-      errorToBool(MaybeRemark.takeError())); // Check for parsing errors.
-  EXPECT_TRUE(*MaybeRemark != nullptr);      // At least one remark.
-
-  const remarks::Remark &Remark = **MaybeRemark;
-  EXPECT_EQ(Remark.RemarkType, remarks::Type::Missed);
-  EXPECT_EQ(checkStr(Remark.PassName, 6), "inline");
-  EXPECT_EQ(checkStr(Remark.RemarkName, 12), "NoDefinition");
-  EXPECT_EQ(checkStr(Remark.FunctionName, 3), "foo");
-  EXPECT_TRUE(Remark.Loc);
-  const remarks::RemarkLocation &RL = *Remark.Loc;
-  EXPECT_EQ(checkStr(RL.SourceFilePath, 6), "file.c");
-  EXPECT_EQ(RL.SourceLine, 3U);
-  EXPECT_EQ(RL.SourceColumn, 12U);
-  EXPECT_TRUE(Remark.Hotness);
-  EXPECT_EQ(*Remark.Hotness, 4U);
-  EXPECT_EQ(Remark.Args.size(), 4U);
-
-  unsigned ArgID = 0;
-  for (const remarks::Argument &Arg : Remark.Args) {
-    switch (ArgID) {
-    case 0:
-      EXPECT_EQ(checkStr(Arg.Key, 6), "Callee");
-      EXPECT_EQ(checkStr(Arg.Val, 3), "bar");
-      EXPECT_FALSE(Arg.Loc);
-      break;
-    case 1:
-      EXPECT_EQ(checkStr(Arg.Key, 6), "String");
-      EXPECT_EQ(checkStr(Arg.Val, 26), " will not be inlined into ");
-      EXPECT_FALSE(Arg.Loc);
-      break;
-    case 2: {
-      EXPECT_EQ(checkStr(Arg.Key, 6), "Caller");
-      EXPECT_EQ(checkStr(Arg.Val, 3), "foo");
-      EXPECT_TRUE(Arg.Loc);
-      const remarks::RemarkLocation &RL = *Arg.Loc;
-      EXPECT_EQ(checkStr(RL.SourceFilePath, 6), "file.c");
-      EXPECT_EQ(RL.SourceLine, 2U);
-      EXPECT_EQ(RL.SourceColumn, 0U);
-      break;
-    }
-    case 3:
-      EXPECT_EQ(checkStr(Arg.Key, 6), "String");
-      EXPECT_EQ(checkStr(Arg.Val, 38),
-                " because its definition is unavailable");
-      EXPECT_FALSE(Arg.Loc);
-      break;
-    default:
-      break;
-    }
-    ++ArgID;
-  }
-
-  MaybeRemark = Parser.next();
-  Error E = MaybeRemark.takeError();
-  EXPECT_TRUE(E.isA<remarks::EndOfFileError>());
-  EXPECT_TRUE(errorToBool(std::move(E))); // Check for parsing errors.
-}
-
-TEST(YAMLRemarks, ParsingBadStringTableIndex) {
-  StringRef Buf = "\n"
-                  "--- !Missed\n"
-                  "Pass: 50\n"
-                  "\n";
-
-  StringRef StrTabBuf = StringRef("inline");
-
-  remarks::ParsedStringTable StrTab(StrTabBuf);
-  Expected<std::unique_ptr<remarks::RemarkParser>> MaybeParser =
-      remarks::createRemarkParser(remarks::Format::YAMLStrTab, Buf,
-                                  std::move(StrTab));
-  EXPECT_FALSE(errorToBool(MaybeParser.takeError()));
-  EXPECT_TRUE(*MaybeParser != nullptr);
-
-  remarks::RemarkParser &Parser = **MaybeParser;
-  Expected<std::unique_ptr<remarks::Remark>> MaybeRemark = Parser.next();
-  EXPECT_FALSE(MaybeRemark); // Expect an error here.
-
-  std::string ErrorStr;
-  raw_string_ostream Stream(ErrorStr);
-  handleAllErrors(MaybeRemark.takeError(),
-                  [&](const ErrorInfoBase &EIB) { EIB.log(Stream); });
-  EXPECT_TRUE(
-      StringRef(Stream.str())
-          .contains("String with index 50 is out of bounds (size = 1)."));
-}
-
 TEST(YAMLRemarks, ParsingGoodMeta) {
   // No metadata should also work.
   parseGoodMeta("--- !Missed\n"
@@ -692,17 +573,6 @@ TEST(YAMLRemarks, ParsingGoodMeta) {
                           "Name: NoDefinition\n"
                           "Function: foo\n",
                           82));
-
-  // Use the string table from the metadata.
-  parseGoodMeta(StringRef("REMARKS\0"
-                          "\0\0\0\0\0\0\0\0"
-                          "\x02\0\0\0\0\0\0\0"
-                          "a\0"
-                          "--- !Missed\n"
-                          "Pass: 0\n"
-                          "Name: 0\n"
-                          "Function: 0\n",
-                          66));
 }
 
 TEST(YAMLRemarks, ParsingBadMeta) {
@@ -727,7 +597,8 @@ TEST(YAMLRemarks, ParsingBadMeta) {
                                  "\0\0\0\0\0\0\0\0"
                                  "\x01\0\0\0\0\0\0\0",
                                  24),
-                       "Expecting string table.", CmpType::Equal);
+                       "String table unsupported for YAML format.",
+                       CmpType::Equal);
 
   parseExpectErrorMeta(StringRef("REMARKS\0"
                                  "\0\0\0\0\0\0\0\0"
